import asyncio
import functools
import inspect
from dataclasses import dataclass
from inspect import (
    Parameter,
    isasyncgenfunction,
    isgeneratorfunction,
)
from typing import (
    Annotated,
    Any,
    Awaitable,
    Callable,
    Self,
    TypeIs,
    TypeVar,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

from starlette.concurrency import run_in_threadpool
from starlette.requests import Request

# from stario.application import Stario
from stario.parameters import RequestParam
from stario.types import CacheScope, RunMode


@dataclass
class Inject:
    func: Callable
    cache: CacheScope = "request"
    mode: RunMode = "auto"


R = TypeVar("R")


class Dependency:

    def __init__(
        self,
        func: Callable[..., Any],
        cache: CacheScope = "request",
        mode: RunMode = "auto",
        return_type: type[Any] | None = None,
        default: Any = Parameter.empty,
    ) -> None:
        # This is what we use to resolve the dependency
        self.func = func
        self.cache = cache
        self.mode = mode
        # Useful info for the tree
        self.is_async = is_async_callable(self.func)
        self.is_gen = is_generator_callable(self.func)
        self.return_type = return_type or get_type_hints(self.func).get("return")
        self.default = default
        self.children: dict[str, Self] = {}

    async def call(self, *args: Any, **kwargs: Any) -> Any:
        if self.mode == "auto":
            # Auto: async if async
            if self.is_async:
                return await self.func(*args, **kwargs)
            elif self.is_gen:
                # This is for a StreamingResponse so we compute it sync
                #   and then it's handled by the StreamingResponse threded / async like it wants
                return self.func(*args, **kwargs)

            # run on threadpool if sync
            return await run_in_threadpool(self.func, *args, **kwargs)

        if self.mode == "sync":
            if self.is_async:
                # We can't run this on threadpool
                return await self.func(*args, **kwargs)

            return self.func(*args, **kwargs)

        if self.mode == "thread":
            return await run_in_threadpool(self.func, *args, **kwargs)

        raise ValueError(f"Unknown run mode: {self.mode}")

    @classmethod
    def _build_function_parameter(cls, param: Parameter) -> Self:
        """
        Based on the function inspect parameter, build a dependency for it.
        """

        if get_origin(param.annotation) is Annotated:

            try:
                # We're interested in the second argument
                # The first is the annotated type, so we don't care
                _, arg, *_ = get_args(param.annotation)

            except ValueError:
                raise ValueError(
                    f"Unknown annotation: {param.name} must be Annotated with one argument"
                )

            if isinstance(arg, RequestParam):
                # Updates url param based on inspect info like defaults etc.

                # Getting those params should be "quick" so we run it sync
                #  rather than delegating to threadpool
                return cls.build(
                    function=arg.digest_parameter(param),
                    mode="sync",
                    default=param.default,
                )

            elif isinstance(arg, Inject):
                return cls.build(
                    function=arg.func,
                    cache=arg.cache,
                    mode=arg.mode,
                    default=param.default,
                )

            return cls.build(function=arg, default=param.default)

        elif isinstance(param.annotation, type) and issubclass(
            param.annotation, Request
        ):
            # This is something of a special case
            return cls(Request)

        elif param.default is not Parameter.empty:
            return cls(
                func=lambda: param.default,
                return_type=param.annotation,  # type: ignore[assignment]
                mode="sync",
                default=param.default,
            )

        else:
            # We cannot really build dependencies for this so we give up
            raise ValueError(
                f"Unknown annotation: {param.name}: {param.annotation} must be Annotated or Request"
            )

    @classmethod
    def build(
        cls,
        function: Callable[..., Any],
        cache: CacheScope = "request",
        mode: RunMode = "auto",
        default: Any = Parameter.empty,
    ) -> Self:
        """
        Builds a tree of dependencies starting from a given function.
        """

        func = get_callable(function)
        signature = inspect.signature(func)
        info = cls(func=func, cache=cache, mode=mode, default=default)

        info.children = {
            param_name: cls._build_function_parameter(param)
            for param_name, param in signature.parameters.items()
            if param_name != "self"
        }

        return info

    async def resolve(
        self,
        request: Request,
        app: Any,
        futures: dict[Callable, Awaitable[Any]],
    ) -> Any:

        if self.func is Request:
            return request

        if self.cache == "app" and self.func in app.cache:
            return app.cache[self.func]

        if self.cache == "request" and self.func in futures:
            return await futures[self.func]

        fut = asyncio.Future()
        if self.cache == "request":
            futures[self.func] = fut

        try:
            if len(self.children) == 1:
                name, child = list(self.children.items())[0]
                arguments = {name: await child.resolve(request, app, futures)}

            elif self.children:
                dep_results = await asyncio.gather(
                    *[d.resolve(request, app, futures) for d in self.children.values()]
                )
                arguments = {k: v for k, v in zip(self.children.keys(), dep_results)}

            else:
                arguments = {}

            # Execute this node's function
            result = await self.call(**arguments)

            if self.cache == "app":
                app.cache[self.func] = result

            elif self.cache == "request":
                fut.set_result(result)

            else:
                # For 'none' scope, return result directly without setting future
                return result  # type: ignore

            return result

        except Exception as e:
            if self.cache in ("request", "app"):
                fut.set_exception(e)
            raise e


T = TypeVar("T")
AwaitableCallable = Callable[..., Awaitable[T]]


@overload
def is_async_callable(obj: AwaitableCallable[T]) -> TypeIs[AwaitableCallable[T]]: ...


@overload
def is_async_callable(obj: Any) -> TypeIs[AwaitableCallable[Any]]: ...


def is_async_callable(obj: Any) -> Any:
    while isinstance(obj, functools.partial):
        obj = obj.func

    return inspect.iscoroutinefunction(obj) or (
        callable(obj) and inspect.iscoroutinefunction(obj.__call__)
    )


def is_generator_callable(obj: Any) -> bool:
    while isinstance(obj, functools.partial):
        obj = obj.func

    return isgeneratorfunction(obj) or isasyncgenfunction(obj)


def get_callable(obj: Any) -> Callable:
    # If it's the function use it, if it's an object with a __call__ use that

    if inspect.isroutine(obj) or inspect.iscoroutinefunction(obj):
        return obj

    return obj.__call__



from typing import Any, AsyncGenerator, Callable, Generator, Protocol

from starlette.requests import Request
from starlette.responses import HTMLResponse, Response, StreamingResponse

# from stario.application import Stario
from stario.datastar import _HtmlProvider, patch_to_sse
from stario.dependencies import Dependency
from stario.types import CacheScope, RunMode


class HandlerWrapper(Protocol):

    def __init__(self, handler: Callable, **kwargs: Any): ...

    async def __call__(self, request: Request) -> Response: ...


class QuickRouteWrapper:

    def __init__(
        self,
        handler: Callable,
        *,
        cache: CacheScope = "request",
        mode: RunMode = "auto",
    ) -> None:
        self.handler = handler
        self.dep = Dependency.build(handler, cache, mode)

    async def __call__(self, request: Request) -> Response:

        app = request.app

        content = await self.dep.resolve(request, app, {})

        if isinstance(content, Response):
            return content

        if isinstance(content, Generator):
            # fmt: off
            return StreamingResponse(
                content    = (patch_to_sse(item) for item in content),
                media_type = "text/event-stream",
                headers = {
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    "X-Accel-Buffering": "no",
                },
            )
            # fmt: on

        if isinstance(content, AsyncGenerator):
            # fmt: off
            return StreamingResponse(
                content    = (patch_to_sse(item) async for item in content),
                media_type = "text/event-stream",
                headers = {
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    "X-Accel-Buffering": "no",
                },
            )
            # fmt: on

        if isinstance(content, _HtmlProvider):
            content = content.__html__()

        return HTMLResponse(
            content,
            status_code=200 if content else 204,
        )



from .application import Stario
from .datastar import LOAD_DATASTAR, Signals
from .dependencies import Inject
from .middlewares import BrotliMiddleware
from .parameters import Body, Cookie, Header, PathParam, QueryParam
from .routes import Command, Query
from .routing import StarRoute
from .types import CacheScope, RunMode

__all__ = [
    "Stario",
    "StarRoute",
    "Query",
    "Command",
    "PathParam",
    "QueryParam",
    "Body",
    "Header",
    "Cookie",
    "Inject",
    "Signals",
    "LOAD_DATASTAR",
    "BrotliMiddleware",
    "CacheScope",
    "RunMode",
]



from typing import Any, Callable, Collection, Mapping, Sequence, TypeVar, override

from starlette.datastructures import Headers, URLPath
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.responses import PlainTextResponse
from starlette.routing import BaseRoute, Match, Route, Router
from starlette.types import ASGIApp, Lifespan, Receive, Scope, Send

type HeadersConstraint = Mapping[str, str | None] | Sequence[
    str | tuple[str, str | None]
]


class StarRoute(Route):

    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        methods: Collection[str] | None = None,
        name: str | None = None,
        include_in_schema: bool = True,
        middleware: Sequence[Middleware] | None = None,
        # Stario specific
        headers: HeadersConstraint | None = None,
        # wrapper: type[HandlerWrapper] = QuickRouteWrapper,
    ) -> None:

        # fmt: off
        super().__init__(
            path              = path,
            endpoint          = endpoint,
            methods           = methods,
            name              = name,
            include_in_schema = include_in_schema,
            middleware        = middleware,
        )
        # fmt: on

        if headers is None:
            self.headers = {}
        elif isinstance(headers, Mapping):
            self.headers = dict(headers)
        else:
            self.headers = {}
            for h in headers:

                if isinstance(h, tuple):
                    self.headers[h[0]] = h[1]
                else:
                    self.headers[h] = None

        # self.wrapper = wrapper

    def _headers_match(self, headers: Headers) -> bool:
        for k, v in self.headers.items():

            if k not in headers:
                return False

            if v is not None and headers.get(k) != v:
                return False

        return True

    @override
    def matches(self, scope: Scope) -> tuple[Match, Scope]:
        # We override this only because we want to support headers constraint

        base_match, base_scope = super().matches(scope)

        # This would fail anyways so we can just return here
        if not self.headers or base_match != Match.FULL:
            return base_match, base_scope

        # I just hope this is light enough so we can create this over and over again
        headers = Headers(scope=scope)
        if not self._headers_match(headers):
            return Match.PARTIAL, base_scope

        # If it's a match anyways, return it :)
        return base_match, base_scope

    @override
    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:

        # If there's nothing to handle additionally, just call the super
        if not self.headers:
            await super().handle(scope, receive, send)
            return

        # If there's something to handle, we need to check the headers
        headers = Headers(scope=scope)
        if not self._headers_match(headers):

            headers_str = ", ".join(str(h) for h in self.headers)
            msg = f"Expected the following headers to be present: {headers_str}"

            if "app" in scope:
                raise HTTPException(status_code=400, detail=msg)
            else:
                response = PlainTextResponse(msg, status_code=400)
            await response(scope, receive, send)
            return

        # All good, call the super
        await super().handle(scope, receive, send)

    @override
    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and list(self.headers) == list(other.headers)

    @override
    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        methods = sorted(self.methods or [])
        path, name = self.path, self.name
        return f"{class_name}(path={path!r}, name={name!r}, methods={methods!r}, headers={self.headers!r})"


_T = TypeVar("_T")


class _DefaultLifespan:
    def __init__(self, router: "StarRouter"):
        self._router = router

    async def __aenter__(self) -> None:
        await self._router.startup()

    async def __aexit__(self, *exc_info: object) -> None:
        await self._router.shutdown()

    def __call__(self: _T, app: object) -> _T:
        return self


class StarRouter:
    def __init__(
        self,
        *routes: BaseRoute,
        redirect_slashes: bool = True,
        default: ASGIApp | None = None,
        lifespan: Lifespan[Any] | None = None,
        middleware: Sequence[Middleware] | None = None,
    ) -> None:

        # This is basically borrowed from Starlette's Router
        self.routes = list(routes)
        self.redirect_slashes = redirect_slashes
        self.default = self.not_found if default is None else default
        self.on_startup = []
        self.on_shutdown = []

        self.lifespan_context = lifespan or _DefaultLifespan(self)

        self.middleware_stack = self.app
        if middleware:
            for cls, args, kwargs in reversed(middleware):
                self.middleware_stack = cls(self.middleware_stack, *args, **kwargs)

    async def not_found(self, scope: Scope, receive: Receive, send: Send) -> None:
        await Router.not_found(self, scope, receive, send)  # type: ignore[arg-type]

    def url_path_for(self, name: str, /, **path_params: Any) -> URLPath:
        return Router.url_path_for(self, name, **path_params)  # type: ignore[arg-type]

    async def startup(self) -> None:
        await Router.startup(self)  # type: ignore[arg-type]

    async def shutdown(self) -> None:
        await Router.shutdown(self)  # type: ignore[arg-type]

    async def lifespan(self, scope: Scope, receive: Receive, send: Send) -> None:
        await Router.lifespan(self, scope, receive, send)  # type: ignore[arg-type]

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        await self.middleware_stack(scope, receive, send)

    async def app(self, scope: Scope, receive: Receive, send: Send) -> None:
        await Router.app(self, scope, receive, send)  # type: ignore[arg-type]

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, StarRouter) and self.routes == other.routes

    def mount(self, path: str, app: ASGIApp, name: str | None = None) -> None:
        Router.mount(self, path, app=app, name=name)  # type: ignore[arg-type]

    def host(self, host: str, app: ASGIApp, name: str | None = None) -> None:
        Router.host(self, host, app=app, name=name)  # type: ignore[arg-type]

    def add(self, route: BaseRoute) -> None:
        self.routes.append(route)

    def add_event_handler(self, event_type: str, func: Callable[[], Any]) -> None:
        Router.add_event_handler(self, event_type, func)  # type: ignore[arg-type]



import json
from dataclasses import dataclass
from typing import Annotated, AsyncGenerator, Literal, Protocol, runtime_checkable

from pydantic import TypeAdapter
from starlette.requests import Request


@runtime_checkable
class _HtmlProvider(Protocol):
    """A type that produces text ready to be placed in an HTML document.

    This is a convention used by html producing/consuming libraries. This lets
    e.g. fasthtml fasttags, or htpy elements, be passed straight in to
    merge_fragments.
    """

    def __html__(self) -> str: ...


@runtime_checkable
class _SSEProvider(Protocol):
    """
    A type that can be converted to a SSE string.
    """

    def to_sse(self) -> str: ...


type HtmlElement = str | _HtmlProvider
type HtmlElements = list[HtmlElement] | HtmlElement
type PatchMode = Literal[
    "outer",
    "inner",
    "replace",
    "prepend",
    "append",
    "before",
    "after",
    "remove",
]
type PatchSelector = str
type SignalValue = str | int | float | bool | list[SignalValue] | dict[str, SignalValue]
type SignalsDict = dict[str, SignalValue]

# SSE related
type PatchSignalsEvent = (PatchSignals | SignalsDict | tuple[SignalsDict, bool])
type PatchElementsEvent = (
    PatchElements
    | HtmlElement
    | tuple[PatchMode, PatchSelector, HtmlElement]
    | tuple[PatchMode, HtmlElement]
    | tuple[Literal["remove"], PatchSelector]
)
type PatchEvent = PatchElementsEvent | PatchSignalsEvent | ExecuteScript | Redirect
type PatchGenerator = AsyncGenerator[PatchEvent, None]


@dataclass(slots=True)
class PatchElements:
    """
    https://data-star.dev/reference/sse_events#datastar-patch-elements
    """

    mode: PatchMode = "outer"
    selector: PatchSelector | None = None
    elements: HtmlElements | None = None
    use_view_transition: bool = False

    # SSE specific
    event_id: str | None = None
    retry_duration: int | None = None

    def to_sse(self) -> str:
        # Standard SSE headers
        lines = ["event: datastar-patch-elements"]

        if self.event_id:
            lines.append(f"id: {self.event_id}")

        if self.retry_duration and self.retry_duration != 1000:
            lines.append(f"retry: {self.retry_duration}")

        # Datastar specific:
        if self.mode and self.mode != "outer":
            lines.append(f"data: mode {self.mode}")

        if self.selector:
            lines.append(f"data: selector {self.selector}")

        if self.use_view_transition:
            lines.append(f"data: useViewTransition {self.use_view_transition}")

        if self.elements:

            if not isinstance(self.elements, list):
                elements = [self.elements]
            else:
                elements = self.elements

            for element in elements:

                if isinstance(element, _HtmlProvider):
                    element = element.__html__()

                # Split elements into lines - this should be faster than splitlines()
                start = 0
                while True:
                    end = element.find("\n", start)
                    if end == -1:
                        lines.append(f"data: elements {element[start:]}")
                        break
                    lines.append(f"data: elements {element[start:end]}")
                    start = end + 1

        return "\n".join(lines) + "\n\n"


@dataclass(slots=True)
class PatchSignals:
    """
    https://data-star.dev/reference/sse_events#datastar-patch-signals
    """

    signals: SignalsDict
    only_if_missing: bool = False

    # SSE specific
    event_id: str | None = None
    retry_duration: int | None = None

    def to_sse(self) -> str:

        # Standard SSE headers
        lines = ["event: datastar-patch-signals"]

        if self.event_id:
            lines.append(f"id: {self.event_id}")

        if self.retry_duration and self.retry_duration != 1000:
            lines.append(f"retry: {self.retry_duration}")

        if self.only_if_missing:
            js_bool = "true" if self.only_if_missing else "false"
            lines.append(f"data: onlyIfMissing {js_bool}")

        # Datastar specific:
        # Split json into lines - this should be faster than splitlines()
        json_str = json.dumps(self.signals)
        start = 0
        while True:
            end = json_str.find("\n", start)
            if end == -1:
                lines.append(f"data: signals {json_str[start:]}")
                break
            lines.append(f"data: signals {json_str[start:end]}")
            start = end + 1

        return "\n".join(lines) + "\n\n"


@dataclass(slots=True)
class RemoveElements:

    selector: PatchSelector | None = None
    elements: HtmlElements | None = None

    def to_sse(self) -> str:
        return PatchElements(
            mode="remove",
            selector=self.selector,
            elements=self.elements,
        ).to_sse()


@dataclass(slots=True)
class ExecuteScript:

    script: str
    auto_remove: bool = True
    # attributes: Mapping[str, str] | list[str] | None = None

    def to_sse(self) -> str:
        attribute_string = ""
        if self.auto_remove:
            attribute_string += ' data-effect="el.remove()"'

        return PatchElements(
            mode="append",
            selector="body",
            elements=f"<script{attribute_string}>{self.script}</script>",
        ).to_sse()


@dataclass(slots=True)
class Redirect:

    location: str

    def to_sse(self) -> str:
        script = f"setTimeout(() => window.location = '{self.location}')"
        return ExecuteScript(script=script).to_sse()


# TODO: Attribute builder from datastar_py

# Datastar quick responses for streaming over sse


def patch_to_sse(patch: PatchEvent) -> str:

    if isinstance(patch, _SSEProvider):
        return patch.to_sse()

    if isinstance(patch, dict):
        return PatchSignals(signals=patch).to_sse()

    if isinstance(patch, (str, _HtmlProvider)):
        return PatchElements(elements=[patch]).to_sse()

    if isinstance(patch, tuple):
        if len(patch) == 3:
            return PatchElements(
                mode=patch[0],
                selector=patch[1],
                elements=patch[2],
            ).to_sse()

        if len(patch) == 2:

            first, second = patch

            if isinstance(first, dict):
                return PatchSignals(
                    signals=first, only_if_missing=bool(second)
                ).to_sse()

            if first != "remove":
                # Accept strings or HTML providers. Disallow bool which can be confused with int.
                if isinstance(second, bool):
                    raise ValueError(f"Unknown patch tuple: {patch}")
                return PatchElements(mode=first, elements=second).to_sse()

            if isinstance(second, str):
                if second.startswith("<") and second.endswith(">"):
                    return PatchElements(mode=first, elements=second).to_sse()

                return PatchElements(mode=first, selector=second).to_sse()

            if isinstance(second, bool):
                raise ValueError(f"Unknown patch tuple: {patch}")

            return PatchElements(mode=first, elements=second).to_sse()

        raise ValueError(f"Unknown patch tuple: {patch}")

    raise ValueError(f"Unknown patch type: {type(patch)}")


LOAD_DATASTAR = '<script type="module" src="https://cdn.jsdelivr.net/gh/starfederation/datastar@main/bundles/datastar.js"></script>'


class SignalsParam:

    def __init__(self, name: str = "datastar"):
        self.name = name
        self.adapter = TypeAdapter(SignalsDict)

    async def __call__(self, request: Request) -> SignalsDict:
        if request.method == "GET":
            raw = request.query_params.get(self.name)
        else:
            raw = await request.body()

        if raw is None:
            return {}

        return self.adapter.validate_json(raw)


type Signals = Annotated[SignalsDict, SignalsParam()]



from typing import Any, Callable, Collection, Sequence

from starlette.middleware import Middleware
from starlette.routing import get_name

from stario.routing import HeadersConstraint, StarRoute
from stario.types import CacheScope, RunMode
from stario.wrappers import QuickRouteWrapper


class Query(StarRoute):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        methods: Collection[str] | None = ["GET"],
        name: str | None = None,
        include_in_schema: bool = True,
        middleware: Sequence[Middleware] | None = None,
        # Stario specific
        headers: HeadersConstraint | None = None,
        cache: CacheScope = "request",
        mode: RunMode = "auto",
    ) -> None:
        # fmt: off
        super().__init__(
            path              = path,
            endpoint          = QuickRouteWrapper(endpoint, cache=cache, mode=mode).__call__,
            methods           = methods,
            name              = name or get_name(endpoint),
            include_in_schema = include_in_schema,
            middleware        = middleware,
            headers           = headers,
        )
        # fmt: on


class Command(StarRoute):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        methods: Collection[str] | None = ["POST"],
        name: str | None = None,
        include_in_schema: bool = True,
        middleware: Sequence[Middleware] | None = None,
        # Stario specific
        headers: HeadersConstraint | None = None,
        cache: CacheScope = "request",
        mode: RunMode = "auto",
    ) -> None:

        # fmt: off
        super().__init__(
            path              = path,
            endpoint          = QuickRouteWrapper(endpoint, cache=cache, mode=mode).__call__,
            methods           = methods,
            name              = name or get_name(endpoint),
            include_in_schema = include_in_schema,
            middleware        = middleware,
            headers           = headers,
        )
        # fmt: on


class CommandDetached(StarRoute):
    pass



from stario.datastar import LOAD_DATASTAR


def ToyInspector() -> str:
    """
    We simply add a div positioned absolutely on the top right of the page
    with the label of key-binding opening the debug panel (CMD+P by default)
    and the content of the debug panel being a pre tag with the attribute
    data-json-signals.

    Should be slightly reduced opacity.

    <pre data-json-signals></pre>
    https://data-star.dev/reference/attributes#data-json-signals
    """

    # Define simple CSS for the debug panel
    debug_panel_css = """
<style>
.stario-debug-panel {
    position: absolute;
    top: 1rem;
    right: 1rem;
    opacity: 0.95;
    border: 1px solid #ccc;
    background: #fff;
    padding: 0.75rem;
    min-width: 220px;
    z-index: 1000;
}
.stario-debug-pre {
    background: #f4f4f4;
    border: 1px solid #eee;
    padding: 0.5rem;
    margin-bottom: 0.25rem;
    font-size: 0.95em;
    max-height: 200px;
    overflow: auto;
}
</style>
"""

    debug_panel_html = """
<div class="stario-debug-panel">
    <b>Debug Inspector:</b>
    <pre class="stario-debug-pre" data-json-signals></pre>
</div>
"""

    return f"""{debug_panel_css}{debug_panel_html}"""


def ToyPage(contents: str, title: str = "Playground") -> str:
    return f"""
<html>
    <head>
        <title>{title}</title>
        {LOAD_DATASTAR}
    </head>
    <body>
        {contents}
        {ToyInspector()}
    </body>
</html>
"""



from inspect import Parameter
from types import UnionType
from typing import Any, Self, cast, get_args, get_origin

from pydantic import TypeAdapter
from starlette.requests import Request


class RequestParam[T]:

    def __init__(self, name: str | None = None) -> None:
        """ """

        self.name: str | None = name
        self.return_type: type[T] | None = None
        self.default: T | object = Parameter.empty
        self.adapter: TypeAdapter[Any] | None = None

    @classmethod
    def full_build(
        cls,
        name: str | None = None,
        return_type: type[T] | None = None,
        default: T | object = Parameter.empty,
    ) -> Self:
        p = cls(name)
        p.return_type = return_type
        p.default = default
        p.adapter = TypeAdapter(return_type) if return_type else None
        return p

    def digest_parameter(self, param: Parameter) -> "RequestParam[T]":
        """
        We assume this will be called after the parameter has been digested.
        So we can use the return type to get the adapter.
        """

        target_type, *_ = get_args(param.annotation)

        self.name = self.name or param.name
        self.default = self.default or param.default
        self.return_type = self.return_type or target_type
        self.adapter = TypeAdapter(self.return_type) if self.return_type else None

        return self

    def extract(self, request: Request) -> Any:
        raise NotImplementedError

    def __call__(self, request: Request) -> T:
        try:
            raw = self.extract(request)
            assert self.adapter is not None
            return self.adapter.validate_python(raw)  # type: ignore[no-any-return]

        except KeyError:
            if self.default is not Parameter.empty:
                return cast(T, self.default)
            raise


class QuerySingle[T](RequestParam[T]):

    def extract(self, request: Request) -> str:
        assert self.name is not None
        return request.query_params[self.name]


class QueryList[T](RequestParam[T]):

    def extract(self, request: Request) -> list[str]:
        assert self.name is not None
        return request.query_params.getlist(self.name)


class QueryParam[T](RequestParam[T]):

    def digest_parameter(self, param: Parameter) -> "RequestParam[T]":

        p = super().digest_parameter(param)

        # we need to decide if user wants a single query param or a list of query params

        if p.return_type is not None and get_origin(p.return_type) is UnionType:

            for union_type in get_args(p.return_type):
                if get_origin(union_type) is list:
                    return QueryList[T].full_build(p.name, union_type, p.default)

        else:
            if p.return_type is not None and get_origin(p.return_type) is list:
                return QueryList[T].full_build(p.name, p.return_type, p.default)

        return QuerySingle[T].full_build(p.name, p.return_type, p.default)


class Body[T]:
    pass


class PathParam[T](RequestParam[T]):

    def extract(self, request: Request) -> str:
        assert self.name is not None
        return request.path_params[self.name]


class Header[T](RequestParam[T]):

    def extract(self, request: Request) -> str:
        assert self.name is not None
        return request.headers[self.name]


class Cookie[T](RequestParam[T]):

    def extract(self, request: Request) -> str:
        assert self.name is not None
        return request.cookies[self.name]
